
#define WATCHDOG

#define MAIN_DELAY_MS						  5	// needed by usb stack
#define READOUT_INTERVAL_MS				  60000	// readout every minute
#define READOUT_WAIT_MS		DS18B20_TCONV_12BIT	// wait long enough for sensors

// USB command codes
#define USBTEMP_CMD_TEST			0
#define USBTEMP_CMD_NO_SENSORS		1
#define USBTEMP_CMD_QUERY_SENSOR	2
#define USBTEMP_CMD_GET_TEMP		3
#define USBTEMP_CMD_SEND			4

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#ifdef WATCHDOG
	#include <avr/wdt.h>
#endif

#include "onewire.h"
#include "ds18x20.h"

#include "usbdrv.h"
#include "delay.h"

typedef struct temp_reading	// internal format for temperature readings
{	uint8_t subzero;
	uint8_t cel;
	uint8_t cel_frac_bits;
} temp_reading_t;

// states of the state machine in main
enum read_state_t {IDLE, MEASURING, UPDATED};

// Temperature globals.
#define MAXSENSORS 5
uint8_t gSensorIDs[MAXSENSORS][OW_ROMCODE_SIZE];
uint8_t nSensors; // number of attached sensors
temp_reading_t readings[MAXSENSORS];

//*****************************************************************************
// Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
// derived from the 66 MHz peripheral clock by dividing. Our timing reference
// is the Start Of Frame signal (a single SE0 bit) available immediately after
// a USB RESET.
//*****************************************************************************
void hadUsbReset(void)
{	const int targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);
	uchar	step = 128;
	uchar	trialValue = 0;
	int		diff, mdiff, pdiff;
//-----------------------------------------------------------------------------
	PORTB &= ~2;	// reset of both outputs
	PORTB &= ~8;	// reset of both outputs

#ifdef WATCHDOG
	wdt_disable ();
#endif

	do	// do a binary search:
	{	OSCCAL = trialValue + step;
		diff = targetValue - usbMeasureFrameLength ();

		if (diff >= 0)
		{	trialValue += step;
		}
		else
		{	diff = -diff;
		}

		step >>= 1;

	} while(step > 0);

	// Find best value (+/-0 is already known)
	OSCCAL++;
	pdiff = targetValue - usbMeasureFrameLength();
	OSCCAL -= 2;
	mdiff = usbMeasureFrameLength() - targetValue;

	// Is diff better than pdiff? (This means diff or mdiff is best)
	if (diff < pdiff)	// Is diff the best? If not: mdiff is already active.
	{	if (diff < mdiff)
		{	OSCCAL = trialValue;
		}
	}
	else
	{	// diff wasn't the best match, choose between mdiff an pdiff.
		// If mdiff is the best match: do nothing since it is already active.
		if (pdiff < mdiff)	// pdiff was the best match.
		{	OSCCAL = trialValue+1;
		}
	}

#ifdef WATCHDOG
	wdt_enable (WDTO_1S);
	WDTCR = 29;	// 8 sek ??
#endif
}
//*****************************************************************************
USB_PUBLIC uchar usbFunctionSetup (uchar data[8])
{	usbRequest_t    *rq = (void *)data;
	static uchar    replyBuf[64];
//-----------------------------------------------------------------------------
	usbMsgPtr = replyBuf;

	if(rq->bRequest == USBTEMP_CMD_TEST)  // ECHO
	{	replyBuf[0] = rq->wValue.bytes[0];
		replyBuf[1] = rq->wValue.bytes[1];

		return 2;
	}

	if(rq->bRequest == USBTEMP_CMD_NO_SENSORS)  // GET_NO_SENSORS -> result = 1 bytes
	{	replyBuf[0] = nSensors;
		return 1;
	}

	// Return id of sensor, 6 bytes.  The first byte contains the type
	// of the sensor.
	if(rq->bRequest == USBTEMP_CMD_QUERY_SENSOR)
	{	uint8_t sensorId=rq->wValue.bytes[0];
	//---------------------------------------
		replyBuf[0] = gSensorIDs[sensorId][0];
		replyBuf[1] = gSensorIDs[sensorId][1];
		replyBuf[2] = gSensorIDs[sensorId][2];
		replyBuf[3] = gSensorIDs[sensorId][3];
		replyBuf[4] = gSensorIDs[sensorId][4];
		replyBuf[5] = gSensorIDs[sensorId][5];
		replyBuf[6] = gSensorIDs[sensorId][6];

		return (7);
	}

	// Return temperature reading of the sensor. The sensor ID is given
	// in wValue. First byte: sensorID, next three bytes: temp. reading.
	if(rq->bRequest == USBTEMP_CMD_GET_TEMP)
	{	uint8_t sensorId=rq->wValue.bytes[0];
	//---------------------------------------
		if (sensorId > nSensors)
			sensorId=0;

		replyBuf[0] = sensorId;
		replyBuf[1] = readings[sensorId].subzero;
		replyBuf[2] = readings[sensorId].cel;
		replyBuf[3] = readings[sensorId].cel_frac_bits;

		return (4);
	}

	if(rq->bRequest == USBTEMP_CMD_SEND)	// Set Pin
	{	replyBuf[0] = rq->wValue.bytes[0];

		if (replyBuf[0] & 1)
			PORTB |= 2;
		else
			PORTB &= ~2;

		if (replyBuf[0] & 2)
			PORTB |= 8;
		else
			PORTB &= ~8;

		return (1);
	}

	return (0);
}
//*****************************************************************************
uint8_t search_sensors(void)
{	uint8_t i, id[OW_ROMCODE_SIZE], diff, nSensors;
//-----------------------------------------------------------------------------
	nSensors = 0;

	for (diff = OW_SEARCH_FIRST; diff != OW_LAST_DEVICE && nSensors < MAXSENSORS;)
	{
		DS18X20_find_sensor (&diff, &id[0]);

		if (diff == OW_PRESENCE_ERR)
		{	break;
		}

		if (diff == OW_DATA_ERR)
		{	break;
		}

		for (i = 0; i < OW_ROMCODE_SIZE; i++)
		{	gSensorIDs[nSensors][i] = id[i];
		}

		nSensors++;
	}

	return (nSensors);
}
//*****************************************************************************
void async_finish_read_sensors(void)
{  uint8_t i, subzero, cel, cel_frac_bits;
//-----------------------------------------------------------------------------
	for (i = 0; i < nSensors; i++)
	{	if (DS18X20_read_meas (&gSensorIDs[i][0], &subzero, &cel, &cel_frac_bits) == DS18X20_OK)
		{	readings[i].subzero=subzero;	// Put the reading in the global readings array
			readings[i].cel=cel;
			readings[i].cel_frac_bits=cel_frac_bits;
		}
	}
}
//*****************************************************************************
void sync_read_sensors (void)
{	DS18X20_start_meas (DS18X20_POWER_PARASITE, NULL);
	delay_ms (DS18B20_TCONV_12BIT);
	async_finish_read_sensors ();
}
//*****************************************************************************
int main(void)
{	uint8_t i;
//-----------------------------------------------------------------------------
	DDRB |= 10;	// PB1 und PB3 als Ausgang

	// We fake an USB disconnect by pulling D+ and D- to 0 during reset. This is
	// necessary if we had a watchdog reset or brownout reset to notify the host
	// that it should re-enumerate the device. Otherwise the host's and device's
	// concept of the device-ID would be out of sync.

	usbDeviceDisconnect ();	// enforce re-enumeration, do this while interrupts are disabled!

#ifdef WATCHDOG
	wdt_reset ();
#endif

	i = 250;

	while (--i)	// fake USB disconnect for > 500 ms
	{	delay_ms (2);
	}

#ifdef WATCHDOG
	wdt_reset ();
#endif

	usbDeviceConnect ();
//	TCCR0 = 5;	// set prescaler to 1/1024
	usbInit ();
	sei ();

	delayloop32 (1000000UL);

	nSensors = search_sensors ();	// search for the available sensors.

	sync_read_sensors ();

	// main loop
	uint32_t delay_counter = 0;
	enum read_state_t state = IDLE;

	for (;;)
	{
#ifdef WATCHDOG
		wdt_reset ();
#endif

		usbPoll ();
		delay_ms (MAIN_DELAY_MS);
		delay_counter += MAIN_DELAY_MS;

		// start measurement.
		if (state == IDLE && delay_counter > READOUT_INTERVAL_MS)
		{	DS18X20_start_meas (DS18X20_POWER_PARASITE, NULL);
			delay_counter = 0;
			state = MEASURING;
		}

		// Read the values from the sensors.
		if (state == MEASURING && delay_counter > READOUT_WAIT_MS)
		{	async_finish_read_sensors ();
			delay_counter = 0;
			state = UPDATED;
		}

		if (state == UPDATED)
		{	delay_counter = 0;
			state = IDLE;
		}
	}
}
